import copy
import json
import numpy as np
import os
import random

from idmtools.assets import Asset

from emodpy_hiv.lib.analysis.age_bin import AgeBin


def add_post_channel_config_as_asset(task, channels, reference_data, site_info):
    """
    (1) Construct a post channel config file
    (2) add this file to consfig builder
    :param config_builder:
    :param reference_data:
    :return:
    """
    channel_config = {channel.name: {} for channel in channels}

    # ck4, I don't really like the 0.5 here; not sure what I can do about it for now
    census_year = site_info['census_year'] + 0.5
    census_age_bin = site_info['census_age_bin']

    # construct channel config
    for channel in channels:
        is_population = channel.name == 'Population'

        # filter by channel
        channel_df = reference_data.filter(keep_only=channel.name)

        channel_config[channel.name]["Type"] = channel.type

        # include population scaling year for Population channel, only
        years = list(channel_df.get_years())
        if is_population and census_year not in years:
            years.append(census_year)
        channel_config[channel.name]["Year"] = sorted(years)

        channel_config[channel.name]["Gender"] = channel_df.get_genders()

        # include population scaling age bin for Population channel, only
        age_bins = [AgeBin.from_string(age_bin_str) for age_bin_str in channel_df.get_age_bins()]
        if is_population and census_age_bin not in age_bins:
            age_bins.append(census_age_bin)
        age_bins = sorted(age_bins, key=lambda age_bin: age_bin.start)
        age_bins = [age_bin.to_tuple() for age_bin in age_bins]
        channel_config[channel.name]["AgeBin"] = age_bins

    # save channel config to a file
    temp_dir = os.getcwd()
    temp_file_path = os.path.join(temp_dir, "post_channel_config.json")
    with open(temp_file_path, 'w') as f:
        f.write(json.dumps(channel_config, sort_keys=True, indent=3))

    # add file to config_builder (as an asset file)
    task.common_assets.add_asset(temp_file_path)


def create_calibration_name(site_name, pop_scale_factor, n_replicates, suffix):
    return '%s--bpsf_%s--replicates_%s--%s' % (site_name, pop_scale_factor, n_replicates, suffix)


def compute_num_cores(max_memory_mb):
    from math import ceil
    return ceil(max_memory_mb/8000)


def set_post_processing(task, path):
    task.common_assets.add_asset(Asset(path, relative_path='python'))
    return task


def initialize_config_object(config_obj, control_params, config_setter):
    # Simulation Setup
    config_obj.parameters.Base_Population_Scale_Factor = control_params.base_population_scale_factor
    config_obj.parameters.Base_Year = control_params.base_year
    config_obj.parameters.Enable_Demographics_Reporting = control_params.Enable_Demographics_Reporting
    config_obj.parameters.Individual_Sampling_Type = control_params.Individual_Sampling_Type
    config_obj.parameters.Simulation_Timestep = control_params.simulation_timestep
    config_obj.parameters.Load_Balance_Filename = control_params.Load_Balance_Filename
    config_obj.parameters.Node_Grid_Size = control_params.Node_Grid_Size
    config_obj.parameters.Random_Number_Generator_Policy = control_params.Random_Number_Generator_Policy
    config_obj.parameters.Random_Number_Generator_Type = control_params.Random_Number_Generator_Type
    config_obj.parameters.Simulation_Duration = control_params.simulation_duration

    # TODO: config hacks until schema fixes arrive
    config_obj.parameters.pop("Serialized_Population_Filenames")
    # config.parameters.pop( "Serialization_Time_Steps" )

    config_obj.parameters.Report_HIV_Event_Channels_List = []
    config_obj.parameters.Male_To_Female_Relative_Infectivity_Ages = [] # 15,25,35 ]
    config_obj.parameters.Male_To_Female_Relative_Infectivity_Multipliers = [] # 5, 1, 0.5 ]
    # This one is crazy! :(
    config_obj.parameters.Maternal_Infection_Transmission_Probability = 0

    config_setter(config_obj)

    return config_obj


def map_sample_to_model_input(simulation, sample, parameter_knob_dict, campaign_builder, custom_sample_mapper,
                              random_run_number=True):
    """
    This method is used directly by the scenario-running script. Do not change its name or argument list.
    s is a dict of user parameters that need mapping to actual model parameters
    """
    params_by_name = {d['Name']: d for d in parameter_knob_dict}

    tags = {}

    if random_run_number:
        tags.update(simulation.task.set_parameter('Run_Number', random.randint(0, 65535)))  # Random random number seed

    value = sample.pop('log base infectivity', None)
    if value is not None:
        tags.update(simulation.task.set_parameter('Base_Infectivity', np.exp(value)))

    # Use any provided map-to parameter names
    remaining_parameters = list(sample.keys())
    for parameter in remaining_parameters:
        value = sample[parameter]
        map_to = params_by_name[parameter].get('MapTo', None)
        if map_to is not None:
            simulation.task.set_parameter(map_to, value)
            tags[parameter] = value
            sample.pop(parameter)

    # Now consume parameters that go to the campaign
    remaining_parameters = copy.deepcopy(sample)
    simulation.task.create_campaign_from_callback(builder=campaign_builder,
                                                  params=sample)
    consumed_parameters = {key: value for key, value in remaining_parameters.items() if key not in sample}
    tags.update(consumed_parameters)

    # Run custom sample mapping routine
    custom_tags = custom_sample_mapper(simulation=simulation, sample=sample)
    tags.update(custom_tags)

    # verify all parameters were mapped
    for parameter_name in sample.keys():
        print('UNUSED PARAMETER:', parameter_name)
    if len(sample) > 0:
        raise ValueError('Parameters not consumed/mapped: %s' % ', '.join(sample.keys()))
    return tags


def constrain_sample(sample, custom_sample_constrainer):
    if ('Pr Ex Trns Male LOW' in sample) and ('Pr Ex Trns Male MED' in sample):
        sample['Pr Ex Trns Male LOW'] = min(sample['Pr Ex Trns Male LOW'], sample['Pr Ex Trns Male MED'])

    if ('Pr Ex Trns Fem LOW' in sample) and ('Pr Ex Trns Fem MED' in sample):
        sample['Pr Ex Trns Fem LOW'] = min(sample['Pr Ex Trns Fem LOW'], sample['Pr Ex Trns Fem MED'])

    sample = custom_sample_constrainer(sample=sample)
    return sample
